change prompt
[calc.git] / calc.c
diff --git a/calc.c b/calc.c
index 6414c2df28f98b7b5b8ffdf1574a6abbda7d3960..749280c38b3bd4dad1e8f82606e96bd1b1a7a143 100644 (file)
--- a/calc.c
+++ b/calc.c
@@ -1,39 +1,32 @@
 /* depend: */
 /* cflags: */
/* linker: atoi.o fdprintf.o */
 
-//#include <malloc.h>
+#include <malloc.h>
+#include <readline/readline.h>
+#include <readline/history.h>
 #include <stddef.h>
 #include <stdio.h>
-#include <unistd.h>
+//#include <stdlib.h>
+//#include <unistd.h>
 
 #include "atoi.h"
+#include "debug.h"
 #include "fdprintf.h"
+#include "parser.h"
 
 /* constants */
 
-//#define BUFFER_SIZE 4096
-#define BUFFER_SIZE 256
-
 /* macros */
 
 #define CEIL(x, y) (((x) + (y) - 1) / (y))
 #define MIN(x, y) (((x) < (y)) ? (x) : (y))
 #define MAX(x, y) (((x) > (y)) ? (x) : (y))
 
-/* verbose */
-
-#define ERROR 0
-#define WARNING 1
-#define INFO 2
-#define DEBUG 3
-
-#define VERBOSE(level, statement...) do { if (level <= verbose) { statement; } } while(0)
-
 /* gobal variables */
 
 char *progname = NULL;
-int verbose = 2;
+int precision = 6;
 
 /* help function */
 
@@ -42,6 +35,7 @@ int usage (int ret)
     int fd = ret ? stdfderr : stdfdout;
     fdprintf (fd, "usage: %s\n", progname);
     fdprintf (fd, " -h : help message\n");
+    fdprintf (fd, " -p : precision (%d)\n", precision);
     fdprintf (fd, " -v : verbose level (%d)\n", verbose);
 
     return ret;
@@ -49,11 +43,11 @@ int usage (int ret)
 
 /* main function */
 
-int main (int argc, char *argv[]) 
+int main (int argc, char *argv[])
 {
-    char buffer[BUFFER_SIZE + 1] = {0};
-    char *pt = buffer;
-    int i = 0, j = 0, n;
+    char *buffer;
+    int i = 0;
+    int ret = 0;
 
     /* program name */
 
@@ -77,6 +71,14 @@ int main (int argc, char *argv[])
         }
         char c = arg[1];
         switch (c) {
+        case 'p':
+            arg = (arg[2]) ? arg + 2 : (--argc > 0) ? *(++argv) : NULL;
+            if (arg == NULL) {
+                PRINTERR ("%s: missing precision\n", progname);
+                return usage (1);
+            }
+            precision = atoi (arg);
+            break;
         case 'v':
             arg = (arg[2]) ? arg + 2 : (--argc > 0) ? *(++argv) : NULL;
             if (arg == NULL) {
@@ -91,42 +93,103 @@ int main (int argc, char *argv[])
         }
     }
 
+    /* format */
+    char format[8] = "=> %.f\n";
+    format[4] = '0' + precision;
+
     /* read from input stream */
 
-    while ((n = read (stdfdin, pt, BUFFER_SIZE - (pt - buffer))) != 0) {
-        VERBOSE (DEBUG, PRINTOUT ("read %d bytes\n", n));
-        n += (pt - buffer);
+    while ((buffer = readline ("<= ")) != NULL) {
 
-        /* look for end of line */
-        for (i = 0, j = 0; i < n; i++) {
-            if (buffer[i] == '\n') {
-                buffer[i] = 0;
-                VERBOSE (DEBUG, PRINTOUT ("line(%d): %s\n", j, buffer + j));
-                //fsync (stdfdout);
-                fflush (stdout);
-                j = i + 1;
-            }
+        /* check empty line */
+        if (strlen (buffer) == 0) {
+            free (buffer);
+            continue;
+        } else if (strcmp (buffer, ".") == 0) {
+            return 0;
         }
 
-        /* keep remainding */
-        if (j < n) {
-            for (i = 0; i < n - j; i++) {
-                buffer[i] = buffer[i + j];
-            }
-            pt = buffer + n - j;
-            for (i = n - j; i < BUFFER_SIZE; i++) {
-                buffer[i] = 0;
+        /* add line into history */
+        add_history (buffer);
+        VERBOSE (INFO, PRINTOUT ("line (%d): %s\n", where_history (), buffer));
+        if (where_history () == 10) {
+            HIST_ENTRY *last = remove_history (0);
+            if (last) {
+                free_history_entry (last);
             }
         }
+
+        /* look for end of line */
+        element_t *element = parser (buffer, NULL, 0);
+        if (element == ERROR_OP) {
+            VERBOSE (WARNING, PRINTOUT ("error while parsing: %s\n", buffer));
+            ret = 1;
+               } else {
+            VERBOSE (INFO, print_element (element, 0));
+            PRINTOUT (format, evaluate_element (element, 0));
+            delelement (element);
+            ret = 0;
+        }
+
+        free (buffer);
     }
 
-    return 0;
+    return ret;
 }
 
 // test: calc.exe -h
 // test: calc.exe -h | awk '/usage:/ { rc=1 } END { exit (1-rc) }'
+// test: echo 1 | calc.exe -v3 | grep -q value
 // test: calc.exe -_ 2> /dev/null | awk 'END { if (NR == 0) { exit(0) } else exit (1) }'
 // test: calc.exe -_ 2>&1 | awk '/usage:/ { rc=1 } END { exit (1-rc) }'
-// test: echo "foo\nbar\nfoobar" | calc.exe -v3
+// test: calc.exe error 2>&1 | grep -q 'invalid option'
+// test: calc.exe -p 2>&1 | grep -q 'missing precision'
+// test: calc.exe -v 2>&1 | grep -q 'missing verbose'
+// test: echo "1 + 2" | calc.exe | grep -q '=> 3'
+// test: echo "1 - 2" | calc.exe | grep -q '=> -1'
+// test: echo "2 * 3" | calc.exe | grep -q '=> 6'
+// test: echo "1 / 2" | calc.exe | grep -q '=> 5e-1'
+// test: echo "8 % 3" | calc.exe | grep -q '=> 2'
+// test: echo "-9 % 3.1" | calc.exe | grep -q '=> -2.8'
+// test: echo "2 ^ 3" | calc.exe | grep -q '=> 8'
+// test: echo "1e-1 + 2.34e3" | calc.exe | grep -q '=> 2.3401'
+// test: echo "sqrt (2)" | calc.exe | grep -q '=> 1.414213'
+// test: echo "pow (2, 3)" | calc.exe | grep -q '=> 8'
+// test: echo "cos (2)" | calc.exe | grep -q '=> -4.161468e-1'
+// test: echo "sin (2)" | calc.exe | grep -q '=> 9.092974e-1'
+// test: echo "atan (2)" | calc.exe | grep -q '=> 1.107148'
+// test: echo "exp (2)" | calc.exe | grep -q '=> 7.389056'
+// test: echo "log (2)" | calc.exe | grep -q '=> 6.931471e-1'
+// test: echo "2 + 3 - 4" | calc.exe | grep -q '=> 1'
+// test: echo "1 + cos (2 - 3)" | calc.exe | grep -q '=> 1.54030'
+// test: echo "cos (1 / 2) * 3" | calc.exe | grep -q '=> 2.63274'
+// test: echo "1 + 4 * (2 - 3)" | calc.exe | grep -q '=> -3'
+// test: echo "(2 - 3) / 4" | calc.exe | grep -q '=> -2.5e-1'
+// test: echo "pow (8 - 3, 4 / 3)" | calc.exe | grep -q '=> 8.549879'
+// test: echo "1 + -2" | calc.exe | grep -q '=> -1'
+// test: echo "1 - +2" | calc.exe | grep -q '=> -1'
+// test: echo "-1 + +2" | calc.exe | grep -q '=> 1'
+// test: echo "-1+2" | calc.exe | grep -q '=> 1'
+// test: echo "1-2" | calc.exe | grep -q '=> -1'
+// test: echo "1 * 2 / 3 + 4" | calc.exe | grep -q '=> 4.666666'
+// test: echo "2 ^ 3 * 4 + 5" | calc.exe | grep -q '=> 3.7e1'
+// test: echo "2 + 3 * 4 ^ 5" | calc.exe | grep -q '=> 3.074e3'
+// test: echo "2 ^ 3 * 4 + cos(5/6)" | calc.exe | grep -q '=> 3.267241e1'
+// test: echo "95-6.3*15-1" | calc.exe | grep -q '=> -5e-1'
+// test: echo "95 - 6.3 * 15 - 1" | calc.exe | grep -q '=> -5e-1'
+// test: echo "95-6.3+15" | calc.exe | grep -q '=> 1.037e2'
+// test: echo "-cos (0) + 1" | calc.exe | grep -q '=> 0'
+// test: echo "quit" | calc.exe | grep -q 'bye'
+// test: echo "help" | calc.exe | grep -q 'miscellaneous'
+// test: echo "1 + 2 *" | calc.exe | grep -q 'error'
+// test: echo "* 1 - 2" | calc.exe | grep -q 'error'
+// test: echo "2 + * 3" | calc.exe | grep -q 'error'
+// test: echo "sqrt 2" | calc.exe | grep -q 'error'
+// test: echo "pow (2)" | calc.exe | grep -q 'error'
+// test: echo "1.23456789" | calc.exe -p 3 | grep -q '1\.234'
+// test: echo . | calc.exe
+// test: echo -e '1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1\n1 + 1' | calc.exe | grep -q 6.4e1
+// test: echo -e '-cos (1)\n1 + 1\n1 - 1\n1 * 1\n1 / 1\n3%2\n2^2\nsqrt (2)\ncos (0)\nsin (0)\natan (0)\nlog (1)\nexp (1)\nhelp\nquit' | calc.exe -v 3 | grep -q bye
+// test: echo -e '1 +\n1 -\n1 * 1\n1 /\n3%\n2^\nsqrt ()\ncos ()\nsin ()\natan ()\nlog ()\nexp ()\n1 + (' | calc.exe |grep -c error |xargs test 11 =
 
 /* vim: set ts=4 sw=4 et: */
@@ -1,39 +1,32 @@
 /* depend: */
 /* cflags: */
-/* linker: atoi.o fdprintf.o */
+/* linker: atoi.o debug.o fdprintf.o parser.o -lm -lreadline */
 